use strict;
use warnings;
use Test::More;

# generated by Dist::Zilla::Plugin::Test::PodSpelling 2.007005
use Test::Spelling 0.12;
use Pod::Wordlist;


add_stopwords(<DATA>);
all_pod_files_spelling_ok( qw( bin lib ) );
__DATA__
Alders
Alders'
Alexandr
Alexey
Andrew
AnnoCPAN
Anwar
Aran
Atsushi
Balakrishnan
Barr
Bilodeau
Blake
BrowserDetect
Cebrin
Christopher
Ciornii
Daniel
David
Deltac
Douglas
Doye
Findlay
Firebird
Florian
Fontani
Graham
Gregory
HTTP
Hao
Heiko
Iceweasel
Ihnen
James
Jensen
Jon
KHTML
Kato
Kimball
Kollar
Kurihara
Latypoff
Lee
MFONTANI
Mahdi
Marco
Maros
McGregor
Mohammad
Moise
Nicolas
Oatis
Olaf
Olivier
Oschwald
Paul
Pelzer
Perlover
Peter
Petr
Purkis
Ramanan
Rask
RealPlayer
Rifkin
Ritter
Robin
Ronald
Semel
Smidsrod
Smidsrød
Sorcinelli
Stadie
Surikov
Thom
UserAgent
Uwe
Voelker
WAP
Walsham
Waters
WebKit
Wilson
WinNT
Wu
Yoshiki
Zarei
adm
ahrefs
alexchorny
altavista
andrew
aol
apache
applecoremedia
askjeeves
audrey
avantgo
baidu
ben
bingbot
bluefeet
browsex
cho45
current
dalvik
daniel
davidihnen
doug
ds
dsi
elinks
facebook
firefox
frank
galeon
gbarr
getright
golib
googleadsbot
googleadsense
googlebot
googlebotimage
googlebotnews
googlebotvideo
googlefavicon
googlemobile
goschwald
gregor
hengst
herrmann
hwu
icab
ie
indy
info
infoseek
iopener
ios
ipad
iphone
ipod
ipsagent
james
joatis
joatis3
jon
khtml
kindlefire
konqueror
ktat
lee
lib
linkchecker
linkexchange
lotusnotes
lsaunders
lycos
macosx
macppc
maintainer
maros
merlynkline
mj
mohammad
mozilla
msn
msnmobile
msoffice
nav
navgold
neoplanet
netfront
netscape
nic
notbenh
nutch
obigo
olaf
olivier
paul
perlover
petr
phplib
polaris
ps
psp
pubsub
puf
ramanan90
realplayer
reneeb
rkimball
robin
rubylib
seamonkey
specialarchiver
staroffice
surikov
thom
ucbrowser
userAgent
uwe
viacheslav
vti
wap
webkit
webos
webtv
wget
winnt
winphone
yandex
yandexbot
yandeximages
yeahoffline
ymgsmz
